;------------------------------------------------------------------------------
;
;  TNKernel RTOS port for ARM(c) Cortex-M3(c) core
;
;  Copyright  2004, 2010 Yuri Tiomkin
;  All rights reserved.
;
;  v.2.6
;
;  Assembler:   ARM(c)
;
;
; Permission to use, copy, modify, and distribute this software in source
; and binary forms and its documentation for any purpose and without fee
; is hereby granted, provided that the above copyright notice appear
; in all copies and that both that copyright notice and this permission
; notice appear in supporting documentation.
;
; THIS SOFTWARE IS PROVIDED BY THE YURI TIOMKIN AND CONTRIBUTORS "AS IS" AND
; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
; ARE DISCLAIMED. IN NO EVENT SHALL YURI TIOMKIN OR CONTRIBUTORS BE LIABLE
; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
; SUCH DAMAGE.
;------------------------------------------------------------------------------


        PRESERVE8

  ;-- External references

  IMPORT  curr_run_task
  IMPORT  next_task_to_run
  IMPORT  myrtos_system_state


  ;-- Public functions declared in this file

  EXPORT  switch_context
  EXPORT  start_exe
  EXPORT  PendSV_Handler
  EXPORT  arm_disable_interrupts
  EXPORT  arm_enable_interrupts

 ;--  Interrupt Control State Register Address

ICSR_ADDR        EQU     0xE000ED04

 ;--  pendSV bit in the Interrupt Control State Register

PENDSVSET        EQU     0x10000000

 ;--  System Handlers 12-15 Priority Register Address

PR_12_15_ADDR    EQU     0xE000ED20

 ;--  PRI_14 (PendSV) priority in the System Handlers 12-15 Priority Register Address
 ;--  PendSV priority is minimal (0xFF)

PENDS_VPRIORITY  EQU     0x00FF0000

;----------------------------------------------------------------------------
;
;----------------------------------------------------------------------------
       AREA    |.text|, CODE, READONLY, ALIGN=3
       THUMB
       REQUIRE8
       PRESERVE8

;----------------------------------------------------------------------------
; Interrups not yet enabled
;----------------------------------------------------------------------------
start_exe

       ldr    r1, =PR_12_15_ADDR        ;-- Load the System 12-15 Priority Register (  r1  PR_12_15_ADDR, -    )
       ldr    r0, [r1]                  ;--       r1    r0
       orr    r0, r0, #PENDS_VPRIORITY  ;-- set PRI_14 (PendSV) to 0xFF - minimal ( or c  PENDS_VPRIORITY -  )
       str    r0, [r1]                  ;--    r0     r1 ..    

       ldr    r1, =myrtos_system_state      ;-- Indicate that system has started (  tn_system_state  r1)
       mov    r0, #1                    ;-- 1 -> TN_SYS_STATE_RUNNING ( r0  1)
       strb   r0, [r1]					;--   r0     r1 ..  tn_system_state

;---------------

       ldr    r1, =curr_run_task     ; =tn_next_task_to_run (   r1)
       ldr    r2, [r1]					;--    r2 (      )
       ldr    r0, [r2]                  ;-- in r0 - new task SP (    r0) -  r0     

       ldmia  r0!, {r4-r11}				;--    r4-r11,        r0, r0+4, r0+8  .. 
       msr    PSP, r0					;--   r0  PSP (PSP/MSP -   cortex m3) r0   = PSP
       orr    lr, lr, #0x04             ;-- Force to new process PSP (LR -  ,  or  4)  LR -  

;--- switch_context_exit

       ldr    r1, =ICSR_ADDR            ;-- Trigger PendSV exception (  PENDSV )
       ldr    r0, =PENDSVSET			;--    -
       str    r0, [r1]					;--  pendsv -  10000000   PENDSVSET

       cpsie  I                         ;-- Enable core interrupts (.  -   pendsv  PendSV_Handler)

;--- Never reach this

       b  .
    ;  bx lr

;----------------------------------------------------------------------------
PendSV_Handler

       cpsid  I                          ;   Disable core interrupts

       ldr    r3, =curr_run_task      ;  in R3 - =tn_curr_run_task
       ldr    r1, [r3]                   ;  in R1 - tn_curr_run_task
       ldr    r2, =next_task_to_run
       ldr    r2, [r2]                   ;  in R2 - tn_next_task_to_run
       cmp    r1, r2					 ;--    tn_next_task_to_run  tn_curr_run_task
       beq    exit_context_switch		 ;--    ;        

;-------------------------------------
		
		;mrs r0, control
		
       mrs    r0, psp                    ;  in PSP - process(task) stack pointer (  r0   -   )
       stmdb  r0!, {r4-r11}				 ;    r0, r0+4, ...   r4 - r11 

       str    r0, [r1]                   ;  save own SP in TCB ( r0    r1 -     tn_curr_run_task) 
       str    r2, [r3]                   ;  in r3 - =tn_curr_run_task
       ldr    r0, [r2]                   ;  in r0 - new task SP ( r0 next_task )

       ldmia  r0!, {r4-r11}				 ;--   r0 d r4 - r11   . 
       msr    psp, r0					 ;--  
       orr    lr, lr, #0x04              ;  Force to new process PSP (LR -  ,  or  4)  LR -  

;-------------------------------------
exit_context_switch

       cpsie  I                          ;  enable core int

       bx     lr

;-----------------------------------------------------------------------------
switch_context

     ;---  Just Invoke PendSV exception

       ldr    r1,  =ICSR_ADDR
       ldr    r0,  =PENDSVSET
       str    r0,  [r1]
	   
	   ;cpsie  I                         ;  Enable core interrupts

       bx     lr

;-----------------------------------------------------------------------------
arm_disable_interrupts

     cpsid  I
     bx     lr

; -------------------------------------------------------------------------
arm_enable_interrupts

     cpsie I
     bx   lr


     END


